package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.zip.CityGmlZipEntry;
import de.hft.stuttgart.citydoctor2.zip.CityGmlZipInputStream;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serial;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Reference object for handling of implicit geometries with a library object contained in an external file
 *
 * @author Riegel
 */
public class LibraryObject extends Geometry {

    private static final Logger logger = LogManager.getLogger(LibraryObject.class);
    @Serial
    private static final long serialVersionUID = -50293435187454911L;
    private final String filepath;
    private final ParserConfiguration config;
    private static Map<String, LibraryObject> libraryObjects = new ConcurrentHashMap<>();

    public static LibraryObject of(Path path, ParserConfiguration config) {
        if (libraryObjects.containsKey(path.toString())) {
            return libraryObjects.get(path.toString());
        }
        Geometry protoGeom = parseFile(path, config);
        if (protoGeom == null) {
            return null;
        }
        LibraryObject libOb = new LibraryObject(protoGeom.getType(), protoGeom.getLod(), path, config);
        protoGeom.getPolygons().forEach(libOb::addPolygon);
        libOb.updateEdgesAndVertices();
        libraryObjects.put(path.toString(), libOb);
        return libOb;
    }

    public static LibraryObject of(CityGmlZipEntry entry, ParserConfiguration config) {
        String fileName = entry.getFileName();
        if (libraryObjects.containsKey(fileName)){
            return libraryObjects.get(fileName);
        }
        Geometry protoGeom = parseZipEntry(entry, config);
        if (protoGeom == null) {
            return null;
        }
        LibraryObject libOb = new LibraryObject(protoGeom.getType(), protoGeom.getLod(), Path.of(fileName), config);
        protoGeom.getPolygons().forEach(libOb::addPolygon);
        libOb.updateEdgesAndVertices();
        libraryObjects.put(fileName, libOb);
        return libOb;
    }

    private LibraryObject(GeometryType type, Lod lod, Path path, ParserConfiguration config) {
        super(type, lod);
        this.filepath = path.toString();
        this.config = config;
    }


    private static Geometry parseFile(Path path, ParserConfiguration config) {
        Geometry geo = null;
        if (path.toFile().exists()) {
            try {
                CityGmlParser.gagLogger(true);
                CityDoctorModel model = CityGmlParser.parseCityGmlFile(path.toString(), config);
                List<CityObject> objects = model.createFeatureStream().toList();
                if (objects.isEmpty()) {
                    throw new InvalidGmlFileException("Referenced library-object's gml file does not contain a CityGML object!");
                } else if (objects.size() > 1) {
                    throw new InvalidGmlFileException("Referenced library-object's gml file contains more than one CityGML object!");
                }
                geo = objects.get(0).getHighestLodGeometry();
            } catch (CityGmlParseException e) {
                logger.error(String.format(
                        "Encountered an error while parsing library object %s", path));
                logger.error(e.getStackTrace());
            } catch (InvalidGmlFileException e) {
                logger.error(e.getStackTrace());
            } finally {
                // Failsafe to remove gag should parsing fail
                CityGmlParser.gagLogger(false);
            }
        } else {
            logger.error(String.format("Implicit geometry references non-existing library object file %s.", path));
        }
        return geo;
    }

    private static Geometry parseZipEntry(CityGmlZipEntry entry, ParserConfiguration config) {
        Geometry geom = null;
        try (CityGmlZipInputStream cgis = new CityGmlZipInputStream(entry)) {
            InputStream is = cgis.getInputStream();
        } catch (IOException e) {
            throw new RuntimeException(e);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        return geom;
    }

}
