package de.hft.stuttgart.citydoctor2.zip;

import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

public class CityGmlZipEntry {

    private static final Logger logger = LogManager.getLogger(CityGmlZipEntry.class);

    private final String fileName;
    private CityDoctorModel model;
    private CityGmlZipArchive parentArchive;
    private boolean validated = false;
    private boolean decompressed = false;
    private long size = -1L;
    private static final long MB = 1024 * 1024L;
    private ZipEntryErrorType errorType = null;

    public static CityGmlZipEntry of(ZipEntry entry, ParserConfiguration config){
        CityGmlZipEntry ze = CityGmlZipEntry.register(entry);
        ze.loadEntry(config);
        return ze;
    }

    public void loadEntry(ParserConfiguration config){
        if (decompressed){
            return;
        }
        if (errorType != null){
            logger.warn("Tried loading erroneous CityGmlZipEntry");
        }
        try{
            CityGmlParser.gagLogger(true);
            this.model = CityGmlParser.parseCityGmlZipEntry(this, config);
            this.decompressed = true;
        } catch (CityGmlParseException | InvalidGmlFileException e) {
            logger.error(e);
            this.errorType = ZipEntryErrorType.INVALID_CITY_GML_FILE;
        } catch (IOException e){
            logger.error(e);
            this.errorType = ZipEntryErrorType.IO_ERROR;
        }
    }

    public static CityGmlZipEntry register(ZipEntry entry){
        CityGmlZipEntry cgzEntry = new CityGmlZipEntry(entry, false);
        try{
            if (!cgzEntry.entrySizeWithinMemoryLimits()) {
                cgzEntry.errorType = ZipEntryErrorType.EXCESSIVE_FILESIZE;
            }
        } catch (IOException e){
            logger.error(e);
            cgzEntry.errorType = ZipEntryErrorType.IO_ERROR;
        }
        return cgzEntry;
    }

    private boolean entrySizeWithinMemoryLimits() throws IOException {
        long memoryLimit = (long) Math.ceil(((double) Runtime.getRuntime().maxMemory() / MB)*0.9);
        if (size != -1L) {
            return memoryLimit > size;
        }
        try (CityGmlZipInputStream cgis = new CityGmlZipInputStream(this)){
            long filesize = cgis.getFileSize(memoryLimit);
            if (filesize != -1){
                this.size = filesize;
                return true;
            } else {
                return false;
            }
        } catch (Exception e) {
            throw new IOException(e);
        }

    }

    protected CityGmlZipEntry(ZipEntry entry, boolean decompressed) {
        this.fileName = entry.getName();
        this.model = null;
        this.decompressed = decompressed;
    }

    protected CityGmlZipEntry(String fileName, boolean decompressed) {
        this.fileName = fileName;
        this.model = null;
        this.decompressed = decompressed;
    }

    public void setArchive(CityGmlZipArchive archive){
        parentArchive = archive;
    }

    public CityGmlZipArchive getArchive(){
        return parentArchive;
    }


    public boolean isValidated(){
        return validated;
    }

    public String getFileName() {
        return fileName;
    }

    public ZipEntryErrorType getErrorType() {
        return errorType;
    }

    public CityDoctorModel getModel() {
        return model;
    }

}
